from Bio.Align import PairwiseAligner
import numpy as np
from typing import Dict
import subprocess
import urllib.request
import os
import numpy as np
from biopandas.pdb import PandasPdb
from proteinflow.utils.mmcif_fix import CustomMmcif
import os
from collections import namedtuple
from operator import attrgetter
import requests
import shutil
import warnings


SIDECHAIN_ORDER = {
    "CYS": ["CB", "SG"],
    "ASP": ["CB", "CG", "OD1", "OD2"],
    "SER": ["CB", "OG"],
    "GLN": ["CB", "CG", "CD", "OE1", "NE2"],
    "LYS": ["CB", "CG", "CD", "CE", "NZ"],
    "ILE": ["CB", "CG1", "CG2", "CD1"],
    "PRO": ["CB", "CG", "CD"],
    "THR": ["CB", "OG1", "CG2"],
    "PHE": ["CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
    "ASN": ["CB", "CG", "OD1", "ND2"],
    "GLY": [],
    "HIS": ["CB", "CG", "ND1", "CD2", "CE1", "NE2"],
    "LEU": ["CB", "CG", "CD1", "CD2"],
    "ARG": ["CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"],
    "TRP": ["CB", "CG", "CD1", "CD2", "NE1", "CE2", "CE3", "CZ2", "CZ3", "CH2"],
    "ALA": ["CB"],
    "VAL": ["CB", "CG1", "CG2"],
    "GLU": ["CB", "CG", "CD", "OE1", "OE2"],
    "TYR": ["CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH"],
    "MET": ["CB", "CG", "SD", "CE"],
}
BACKBONE_ORDER = ["N", "C", "CA", "O"]

d3to1 = {
    "CYS": "C",
    "ASP": "D",
    "SER": "S",
    "GLN": "Q",
    "LYS": "K",
    "ILE": "I",
    "PRO": "P",
    "THR": "T",
    "PHE": "F",
    "ASN": "N",
    "GLY": "G",
    "HIS": "H",
    "LEU": "L",
    "ARG": "R",
    "TRP": "W",
    "ALA": "A",
    "VAL": "V",
    "GLU": "E",
    "TYR": "Y",
    "MET": "M",
}
S3Obj = namedtuple("S3Obj", ["key", "mtime", "size", "ETag"])


class PDBError(ValueError):
    pass


def _s3list(
    bucket,
    path,
    start=None,
    end=None,
    recursive=True,
    list_dirs=True,
    list_objs=True,
    limit=None,
):
    """
    Iterator that lists a bucket's objects under path, (optionally) starting with
    start and ending before end.

    If recursive is False, then list only the "depth=0" items (dirs and objects).

    If recursive is True, then list recursively all objects (no dirs).

    Args:
        bucket:
            a boto3.resource('s3').Bucket().
        path:
            a directory in the bucket.
        start:
            optional: start key, inclusive (may be a relative path under path, or
            absolute in the bucket)
        end:
            optional: stop key, exclusive (may be a relative path under path, or
            absolute in the bucket)
        recursive:
            optional, default True. If True, lists only objects. If False, lists
            only depth 0 "directories" and objects.
        list_dirs:
            optional, default True. Has no effect in recursive listing. On
            non-recursive listing, if False, then directories are omitted.
        list_objs:
            optional, default True. If False, then directories are omitted.
        limit:
            optional. If specified, then lists at most this many items.

    Returns:
        an iterator of S3Obj.

    Examples:
        # set up
        >>> s3 = boto3.resource('s3')
        ... bucket = s3.Bucket('bucket-name')

        # iterate through all S3 objects under some dir
        >>> for p in s3list(bucket, 'some/dir'):
        ...     print(p)

        # iterate through up to 20 S3 objects under some dir, starting with foo_0010
        >>> for p in s3list(bucket, 'some/dir', limit=20, start='foo_0010'):
        ...     print(p)

        # non-recursive listing under some dir:
        >>> for p in s3list(bucket, 'some/dir', recursive=False):
        ...     print(p)

        # non-recursive listing under some dir, listing only dirs:
        >>> for p in s3list(bucket, 'some/dir', recursive=False, list_objs=False):
        ...     print(p)
    """

    kwargs = dict()
    if start is not None:
        if not start.startswith(path):
            start = os.path.join(path, start)
    if end is not None:
        if not end.startswith(path):
            end = os.path.join(path, end)
    if not recursive:
        kwargs.update(Delimiter="/")
        if not path.endswith("/") and len(path) > 0:
            path += "/"
    kwargs.update(Prefix=path)
    if limit is not None:
        kwargs.update(PaginationConfig={"MaxItems": limit})

    paginator = bucket.meta.client.get_paginator("list_objects")
    for resp in paginator.paginate(Bucket=bucket.name, **kwargs):
        q = []
        if "CommonPrefixes" in resp and list_dirs:
            q = [S3Obj(f["Prefix"], None, None, None) for f in resp["CommonPrefixes"]]
        if "Contents" in resp and list_objs:
            q += [
                S3Obj(f["Key"], f["LastModified"], f["Size"], f["ETag"])
                for f in resp["Contents"]
            ]
        # note: even with sorted lists, it is faster to sort(a+b)
        # than heapq.merge(a, b) at least up to 10K elements in each list
        q = sorted(q, key=attrgetter("key"))
        if limit is not None:
            q = q[:limit]
            limit -= len(q)
        for p in q:
            if end is not None and p.key >= end:
                return
            yield p


def _retrieve_author_chain(chain):
    """
    Retrieve the (author) chain names present in the chain section (delimited by '|' chars) of a header line in a fasta file
    """

    if "auth" in chain:
        return chain.split(" ")[-1][:-1]

    return chain


def _retrieve_chain_names(entry):
    """
    Retrieve the (author) chain names present in one header line of a fasta file (line that begins with '>')
    """

    entry = entry.split("|")[1]

    if "Chains" in entry:
        return [_retrieve_author_chain(e) for e in entry[7:].split(", ")]

    return [_retrieve_author_chain(entry[6:])]


def _retrieve_fasta_chains(fasta_file):
    """
    Return a dictionary containing all the (author) chains in a fasta file (keys) and their corresponding sequence
    """

    with open(fasta_file, "r") as f:
        lines = np.array(f.readlines())

    indexes = np.array([k for k, l in enumerate(lines) if l[0] == ">"])
    starts = indexes + 1
    ends = list(indexes[1:]) + [len(lines)]
    names = lines[indexes]
    seqs = ["".join(lines[s:e]).replace("\n", "") for s, e in zip(starts, ends)]

    out_dict = {}
    for name, seq in zip(names, seqs):
        for chain in _retrieve_chain_names(name):
            out_dict[chain] = seq

    return out_dict


def _open_structure(file_path: str, tmp_folder: str) -> Dict:
    """
    Read a PDB file and parse it into a dictionary if it meets criteria

    The criteria are:
    - only contains proteins,
    - resolution is known and is not larger than the threshold.

    The output dictionary has the following keys:
    - `'crd_raw'`: a `pandas` (`biopandas`) table with the coordinates,
    - `'fasta'`: a dictionary where keys are chain ids and values are fasta sequences.

    Parameters
    ----------
    file_path : str
        the path to the .pdb.gz file
    tmp_folder : str
        the path to the temporary data folder
    thr_resolution : float, default 3.5
        the resolution threshold

    Output
    ------
    pdb_dict : Dict
        the parsed dictionary
    """

    cif = file_path.endswith("cif.gz")
    pdb, biounit = os.path.basename(file_path).split(".")[0].split("-")
    out_dict = {}

    # download fasta and check if it contains only proteins
    fasta_path = os.path.join(tmp_folder, f"{pdb}.fasta")
    try:
        seqs_dict = _retrieve_fasta_chains(fasta_path)
    except FileNotFoundError:
        raise PDBError("Fasta file not found")

    # load coordinates in a nice format
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if cif:
                p = CustomMmcif().read_mmcif(file_path).get_model(1)
            else:
                p = PandasPdb().read_pdb(file_path).get_model(1)
    except FileNotFoundError:
        raise PDBError("PDB / mmCIF file downloaded but not found")
    out_dict["crd_raw"] = p.df["ATOM"]
    out_dict["seq_df"] = p.amino3to1()

    # # add metadata
    # metadata = parse_pdb_header(file_path)
    # for key in ["structure_method"]:
    #     out_dict[key] = metadata.get(key)

    # retrieve sequences that are relevant for this PDB from the fasta file
    chains = p.df["ATOM"]["chain_id"].unique()

    if not set([x.split("-")[0] for x in chains]).issubset(set(list(seqs_dict.keys()))):
        raise PDBError("Some chains in the PDB do not appear in the fasta file")

    out_dict["fasta"] = {k: seqs_dict[k.split("-")[0]] for k in chains}

    try:
        os.remove(file_path)
    except OSError:
        pass
    return out_dict


def _align_structure(
    pdb_dict: Dict,
    min_length: int = 30,
    max_length: int = None,
    max_missing_middle: float = 0.1,
    max_missing_ends: float = 0.3,
) -> Dict:
    """
    Align and filter a PDB dictionary

    The filtering criteria are:
    - only contains natural amino acids,
    - number of non-missing residues per chain is not smaller than `min_length`,
    - fraction of missing residues per chain is not larger than `max_missing`,
    - number of residues per chain is not larger than `max_length` (if provided).

    The output is a nested dictionary where first-level keys are chain Ids and second-level keys are the following:
    - `'crd_bb'`: a `numpy` array of shape `(L, 4, 3)` with backbone atom coordinates (N, Ca, C, O),
    - `'crd_sc'`: a `numpy` array of shape `(L, 10, 3)` with sidechain atom coordinates (in a fixed order),
    - `'msk'`: a `numpy` array of shape `(L,)` where ones correspond to residues with known coordinates and
        zeros to missing values,
    - `'seq'`: a string of length `L` with residue types.

    Parameters
    ----------
    pdb_dict : Dict
        the output of `open_pdb`
    min_length : int, default 30
        the minimum number of non-missing residues per chain
    max_length : int, optional
        the maximum number of residues per chain

    Returns
    -------
    pdb_dict : Dict | None
        the parsed dictionary or `None`, if the criteria are not met
    """

    crd = pdb_dict["crd_raw"]
    fasta = pdb_dict["fasta"]
    seq_df = pdb_dict["seq_df"]
    pdb_dict = {}
    crd = crd[crd["record_name"] == "ATOM"]

    if len(crd["chain_id"].unique()) == 0:
        raise PDBError("No chains found")

    if not crd["residue_name"].isin(d3to1.keys()).all():
        raise PDBError("Unnatural amino acids found")

    for chain in crd["chain_id"].unique():
        pdb_dict[chain] = {}
        chain_crd = crd[crd["chain_id"] == chain].reset_index()

        if len(chain_crd) / len(fasta[chain]) < 1 - (
            max_missing_ends + max_missing_middle
        ):
            raise PDBError("Too many missing values in total")

        # align fasta and pdb and check criteria)
        pdb_seq = "".join(seq_df[seq_df["chain_id"] == chain]["residue_name"])
        if "insertion" in chain_crd.columns:
            chain_crd["residue_number"] = chain_crd.apply(
                lambda row: f"{row['residue_number']}_{row['insertion']}", axis=1
            )
        unique_numbers = chain_crd["residue_number"].unique()
        if len(unique_numbers) != len(pdb_seq):
            raise PDBError("Inconsistencies in the biopandas dataframe")
        aligner = PairwiseAligner()
        aligner.match_score = 2
        aligner.mismatch_score = -10
        aligner.open_gap_score = -0.5
        aligner.extend_gap_score = -0.1
        aligned_seq, fasta_seq = aligner.align(pdb_seq, fasta[chain])[0]
        aligned_seq_arr = np.array(list(aligned_seq))
        if "-" in fasta_seq or "".join([x for x in aligned_seq if x != "-"]) != pdb_seq:
            raise PDBError("Incorrect alignment")
        residue_numbers = np.where(aligned_seq_arr != "-")[0]
        start = residue_numbers.min()
        end = residue_numbers.max() + 1
        if start + (len(aligned_seq) - end) > max_missing_ends * len(aligned_seq):
            raise PDBError("Too many missing values in the ends")
        if (aligned_seq_arr[start:end] == "-").sum() > max_missing_middle * (
            end - start
        ):
            raise PDBError("Too many missing values in the middle")
        pdb_dict[chain]["seq"] = fasta[chain]
        pdb_dict[chain]["msk"] = (aligned_seq_arr != "-").astype(int)
        l = sum(pdb_dict[chain]["msk"])
        if min_length is not None and l < min_length:
            raise PDBError("Sequence is too short")
        if max_length is not None and len(aligned_seq) > max_length:
            raise PDBError("Sequence is too long")

        # go over rows of coordinates
        crd_arr = np.zeros((len(aligned_seq), 14, 3))

        def arr_index(row):
            atom = row["atom_name"]
            if atom.startswith("H") or atom == "OXT":
                return -1  # ignore hydrogens and OXT
            order = BACKBONE_ORDER + SIDECHAIN_ORDER[row["residue_name"]]
            try:
                return order.index(atom)
            except:
                raise PDBError(f"Unexpected atoms ({atom})")

        indices = chain_crd.apply(arr_index, axis=1)
        indices = indices.astype(int)
        informative_mask = indices != -1
        res_indices = np.where(aligned_seq_arr != "-")[0]
        replace_dict = {x: y for x, y in zip(unique_numbers, res_indices)}
        chain_crd["residue_number"].replace(replace_dict, inplace=True)
        crd_arr[
            chain_crd[informative_mask]["residue_number"], indices[informative_mask]
        ] = chain_crd[informative_mask][["x_coord", "y_coord", "z_coord"]]

        pdb_dict[chain]["crd_bb"] = crd_arr[:, :4, :]
        pdb_dict[chain]["crd_sc"] = crd_arr[:, 4:, :]
        pdb_dict[chain]["msk"][(pdb_dict[chain]["crd_bb"] == 0).sum(-1).sum(-1) > 0] = 0
        if (pdb_dict[chain]["msk"][start:end] == 0).sum() > max_missing_middle * (
            end - start
        ):
            raise PDBError("Too many missing values in the middle")
    return pdb_dict



import os
import subprocess
import numpy as np
import pickle as pkl
from tqdm import tqdm
import shutil


def _get_s3_paths_from_tag(tag):
    """
    Get the path to the data and split dictionary folders on S3 given a tag
    """

    dict_path = f"s3://proteinflow-datasets/proteinflow_{tag}_splits_dict/"
    data_path = f"s3://proteinflow-datasets/proteinflow_{tag}/"
    return data_path, dict_path


def _biounits_in_clusters_dict(clusters_dict):
    """
    Return the list of all biounit files present in clusters_dict
    """

    return np.unique([c[0] for c in list(np.concatenate(list(clusters_dict.values())))])


def _download_dataset_dicts_from_s3(
    dict_folder_path, s3_path="s3://ml4-main-storage/proteinflow_20221110_splits_dict/"
):
    """
    Download dictionaries containing database split information from s3 to a local folder
    """

    train_path = os.path.join(s3_path, "train.pickle")
    valid_path = os.path.join(s3_path, "valid.pickle")
    test_path = os.path.join(s3_path, "test.pickle")

    if not os.path.exists(dict_folder_path):
        os.makedirs(dict_folder_path)

    subprocess.run(
        ["aws", "s3", "cp", "--no-sign-request", train_path, dict_folder_path]
    )
    subprocess.run(
        ["aws", "s3", "cp", "--no-sign-request", valid_path, dict_folder_path]
    )
    subprocess.run(
        ["aws", "s3", "cp", "--no-sign-request", test_path, dict_folder_path]
    )


def _split_data(dataset_path="./data/proteinflow_20221110/"):
    """
    Rearrange files into folders according to the dataset split dictionaries at `dataset_path/splits_dict`

    Parameters
    ----------
    dataset_path : str, default "./data/proteinflow_20221110/"
        The path to the dataset folder containing pre-processed entries and a `splits_dict` folder with split dictionaries (downloaded or generated with `get_split_dictionaries`)
    """

    dict_folder = os.path.join(dataset_path, "splits_dict")
    with open(os.path.join(dict_folder, "train.pickle"), "rb") as f:
        train_clusters_dict = pkl.load(f)
    with open(os.path.join(dict_folder, "valid.pickle"), "rb") as f:
        valid_clusters_dict = pkl.load(f)
    with open(os.path.join(dict_folder, "test.pickle"), "rb") as f:
        test_clusters_dict = pkl.load(f)

    train_biounits = _biounits_in_clusters_dict(train_clusters_dict)
    valid_biounits = _biounits_in_clusters_dict(valid_clusters_dict)
    test_biounits = _biounits_in_clusters_dict(test_clusters_dict)
    train_path = os.path.join(dataset_path, "train")
    valid_path = os.path.join(dataset_path, "valid")
    test_path = os.path.join(dataset_path, "test")

    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)

    if not os.path.exists(train_path):
        os.makedirs(train_path)
    if not os.path.exists(valid_path):
        os.makedirs(valid_path)
    if not os.path.exists(test_path):
        os.makedirs(test_path)

    print("Moving files in the train set...")
    for biounit in tqdm(train_biounits):
        shutil.move(os.path.join(dataset_path, biounit), train_path)
    print("Done!")
    print("Moving files in the validation set...")
    for biounit in tqdm(valid_biounits):
        shutil.move(os.path.join(dataset_path, biounit), valid_path)
    print("Done!")
    print("Moving files in the test set...")
    for biounit in tqdm(test_biounits):
        shutil.move(os.path.join(dataset_path, biounit), test_path)
    print("Done!")


def _download_dataset_from_s3(
    dataset_path="./data/proteinflow_20221110/",
    s3_path="s3://ml4-main-storage/proteinflow_20221110/",
):
    """
    Download the pre-processed files
    """

    if s3_path.startswith("s3"):
        print("Downloading the dataset from s3...")
        subprocess.run(
            ["aws", "s3", "sync", "--no-sign-request", s3_path, dataset_path]
        )
        print("Done!")
    else:
        shutil.move(s3_path, dataset_path)


def _download_dataset(tag, local_datasets_folder="./data/"):
    """
    Download the pre-processed data and the split dictionaries

    Parameters
    ----------
    tag : str
        name of the dataset (check `get_available_tags` to see the options)
    local_dataset_folder : str, default "./data/"
        the local folder that will contain proteinflow dataset folders, temporary files and logs

    Returns
    -------
    data_folder : str
        the path to the downloaded data folder
    """

    s3_data_path, s3_dict_path = _get_s3_paths_from_tag(tag)
    data_folder = os.path.join(local_datasets_folder, f"proteinflow_{tag}")
    dict_folder = os.path.join(
        local_datasets_folder, f"proteinflow_{tag}", "splits_dict"
    )

    print("Downloading dictionaries for splitting the dataset...")
    _download_dataset_dicts_from_s3(dict_folder, s3_dict_path)
    print("Done!")

    _download_dataset_from_s3(dataset_path=data_folder, s3_path=s3_data_path)
    return data_folder



import os
import subprocess
import editdistance
import numpy as np
import pickle as pkl
from tqdm import tqdm
from collections import Counter


def _open_pdb(file):
    """
    Open a PDB file in the pickle format that follows the dwnloading and processing of the database
    """

    with open(file, "rb") as f:
        return pkl.load(f)


def _compare_identity(seq, seqs, threshold):
    """
    Assess whether a sequence is in a list of sequences (in the sense that it shares at least 90% to one of the sequences in the list)
    """

    for s in seqs:
        if editdistance.eval(s, seq) / max(len(s), len(seq)) <= (1 - threshold):
            return True

    return False


def _compare_seqs(seqs1, seqs2, threshold):
    """
    Assess whether 2 lists of sequences contain exactly the same set of sequences
    """

    for seq in seqs1:
        if not _compare_identity(seq, seqs1, threshold):
            return False

    for seq in seqs2:
        if not _compare_identity(seq, seqs2, threshold):
            return False

    return True


def _check_biounits(biounits_list, threshold):
    """
    Return the indexes of the redundant biounits within the list of files given by `biounits_list`
    """

    biounits = [_open_pdb(b) for b in biounits_list]
    indexes = []

    for k, b1 in enumerate(biounits):
        if k not in indexes:
            b1_seqs = [b1[chain]["seq"] for chain in b1.keys()]
            for l, b2 in enumerate(biounits[k + 1 :]):
                if len(b1.keys()) != len(b2.keys()):
                    continue

                b2_seqs = [b2[chain]["seq"] for chain in b2.keys()]
                if _compare_seqs(b1_seqs, b2_seqs, threshold):
                    indexes.append(k + l + 1)

    return indexes


def _remove_database_redundancies(dir, seq_identity_threshold=0.9):
    """
    Remove all biounits in the database that are copies to another biounits in terms of sequence

    Sequence identity is definded by the 'seq_identity_threshold' parameter for robust detection of sequence similarity (missing residues, point mutations, ...).

    Parameters
    ----------
    dir : str
        the path to the database where all the biounits are stored in pickle files after their processing
    seq_identity_threshold : float, default .9
        the threshold that determines up to what percentage identity sequences are considered as the same

    Returns
    -------
    total_removed : int
        the total number of removed biounits
    """

    all_files = np.array(os.listdir(dir))
    all_pdbs = np.array([file[:4] for file in all_files])
    pdb_counts = Counter(all_pdbs)
    pdbs_to_check = [pdb for pdb in pdb_counts.keys() if pdb_counts[pdb] > 1]
    total_removed = []

    for pdb in tqdm(pdbs_to_check):
        biounits_list = np.array(
            [os.path.join(dir, file) for file in all_files[all_pdbs == pdb]]
        )
        biounits_list = sorted(biounits_list)
        redundancies = _check_biounits(biounits_list, seq_identity_threshold)
        if redundancies != []:
            for k in redundancies:
                total_removed.append(os.path.basename(biounits_list[k]).split(".")[0])
                subprocess.run(["rm", biounits_list[k]])

    return total_removed



from aiobotocore.session import get_session
from botocore import UNSIGNED
from botocore.config import Config
import asyncio
import multiprocessing as mp
from pathos.multiprocessing import ProcessingPool as Pool
import os
from p_tqdm import p_map


async def _getobj(client, key):
    resp = await client.get_object(Bucket="pdbsnapshots", Key=key)
    return await resp["Body"].read()


async def _download_file(client, snapshots, tmp_folder, id):
    pdb_id, biounit = id.lower().split("-")
    prefixes = [
        "pub/pdb/data/biounit/PDB/all/",
        "pub/pdb/data/biounit/mmCIF/all/",
        "pub/pdb/data/assemblies/mmCIF/all/",
    ]
    types = ["pdb", "cif", "cif"]
    filenames = {
        "cif": f"{pdb_id}-assembly{biounit}.cif.gz",
        "pdb": f"{pdb_id}.pdb{biounit}.gz",
    }
    for folder in snapshots:
        for prefix, t in zip(prefixes, types):
            file = folder + prefix + filenames[t]
            local_path = os.path.join(tmp_folder, f"{pdb_id}-{biounit}") + f".{t}.gz"
            try:
                obj = await _getobj(client, file)
                with open(local_path, "wb") as f:
                    f.write(obj)
                return local_path
            except Exception as e:
                pass
    return id


async def _go(download_list, tmp_folder, snapshots):
    session = get_session()
    async with session.create_client(
        "s3", config=Config(signature_version=UNSIGNED)
    ) as client:
        tasks = [
            _download_file(
                client=client,
                snapshots=snapshots,
                # key=os.path.join(prefix, f"{x.split('-')[0]}.pdb{x.split('-')[1]}.gz"),
                tmp_folder=tmp_folder,
                id=x,
            )
            for x in download_list
        ]
        out = await asyncio.gather(*tasks)
    return out


def _singleProcess(download_list, tmp_folder, snapshots):
    """mission for single process"""
    loop = asyncio.new_event_loop()
    return loop.run_until_complete(
        _go(download_list, tmp_folder=tmp_folder, snapshots=snapshots)
    )


def _download_s3_parallel(pdb_ids, tmp_folder, snapshots):
    # number of process
    no_tasks = max(16, len(pdb_ids) // 5000)

    download_list_chunk = [pdb_ids[i::no_tasks] for i in range(no_tasks)]
    out = p_map(
        lambda x: _singleProcess(
            download_list=x, tmp_folder=tmp_folder, snapshots=snapshots
        ),
        download_list_chunk,
    )
    return out



import os
import copy
import subprocess
import editdistance
import numpy as np
import random as rd
import pickle as pkl
import networkx as nx
from tqdm import tqdm
from collections import defaultdict
from itertools import combinations


def _unique_chains(seqs_list):
    """
    Get unique chains
    """

    new_seqs_list = [seqs_list[0]]
    chains = [new_seqs_list[0][0]]

    for seq in seqs_list[1:]:
        if seq[0] not in chains:
            new_seqs_list.append(seq)
            chains.append(seq[0])

    return new_seqs_list


def _merge_chains(seqs_dict_):
    """
    Look into the chains of each PDB and regroup redundancies (at 90% sequence identity)
    """

    seqs_dict = copy.deepcopy(seqs_dict_)
    pdbs_to_delete = []

    for pdb in tqdm(seqs_dict.keys()):
        if seqs_dict[pdb] == []:
            pdbs_to_delete.append(pdb)
            continue

        seqs_dict[pdb] = _unique_chains(seqs_dict[pdb])
        groups, ref_seqs, indexes = [], [], []

        for k in range(len(seqs_dict[pdb])):
            if k in indexes:
                continue
            group = [seqs_dict[pdb][k][0]]
            ref_seq = seqs_dict[pdb][k][1]
            ref_seqs.append(ref_seq)
            indexes.append(k)

            for l in range(k + 1, len(seqs_dict[pdb])):
                chain, seq = seqs_dict[pdb][l][0], seqs_dict[pdb][l][1]
                if (
                    l in indexes
                    or len(seq) > 1.1 * len(ref_seq)
                    or len(seq) < 0.9 * len(ref_seq)
                    or editdistance.eval(seq, ref_seq) / max(len(seq), len(ref_seq))
                    > 0.1
                ):
                    continue
                group.append(chain)
                indexes.append(l)

            groups.append(group)

        new_group = []
        for group, seq in zip(groups, ref_seqs):
            new_group.append(("-".join(group), seq))
        seqs_dict[pdb] = new_group

    for pdb in pdbs_to_delete:
        del seqs_dict[pdb]

    return seqs_dict


def _load_pdbs(dir):
    """
    Load biounits and group their sequences by PDB and similarity (90%)
    """

    seqs_dict = defaultdict(lambda: [])

    for file in tqdm(os.listdir(dir)):
        load_path = os.path.join(dir, file)
        if os.path.isdir(load_path):
            continue
        with open(load_path, "rb") as f:
            pdb_dict = pkl.load(f)
        seqs = [(chain, pdb_dict[chain]["seq"]) for chain in pdb_dict.keys()]
        seqs_dict[file[:4]] += seqs

    return seqs_dict


def _write_fasta(fasta_path, merged_seqs_dict):
    """
    Write a fasta file containing all the sequences contained in the merged_seqs_dict dictionary
    """

    with open(fasta_path, "w") as f:
        for k in merged_seqs_dict.keys():
            for chain, seq in merged_seqs_dict[k]:
                f.write(">" + k + "_" + chain + "\n")
                f.write(seq + "\n")


def _run_mmseqs2(fasta_file, tmp_folder, min_seq_id):
    """
    Run the MMSeqs2 command with the parameters we want

    Results are stored in the tmp_folder/MMSeqs2 directory.
    """

    os.makedirs(os.path.join(tmp_folder, "MMSeqs2_results"), exist_ok=True)
    subprocess.run(
        [
            "mmseqs",
            "easy-cluster",
            fasta_file,
            os.path.join(tmp_folder, "MMSeqs2_results/clusterRes"),
            os.path.join(tmp_folder, "MMSeqs2_results/tmp"),
            "--min-seq-id",
            str(min_seq_id),
        ]
    )


def _read_clusters(cluster_file_fasta):
    """
    Read the output from MMSeqs2 and produces 2 dictionaries that store the clusters information

    In cluster_dict, keys are the full names (pdb + chains) whereas in cluster_pdb_dict, keys are just the PDB ids (so less clusters but bigger).
    """

    with open(cluster_file_fasta, "r") as f:
        cluster_dict = defaultdict(lambda: [])
        cluster_pdb_dict = defaultdict(lambda: [])
        cluster_name, sequence_name = None, None
        found_header = False

        for line in f.readlines():
            if line[0] == ">" and found_header:
                cluster_name = line[1:-1]
                sequence_name = line[1:-1]

            elif line[0] == ">":
                sequence_name = line[1:-1]
                found_header = True

            else:
                cluster_dict[cluster_name].append(sequence_name)
                cluster_pdb_dict[cluster_name].append(sequence_name[:4])
                found_header = False

        for k in cluster_pdb_dict.keys():
            cluster_pdb_dict[k] = np.unique(cluster_pdb_dict[k])

    return cluster_dict, cluster_pdb_dict


def _make_graph(cluster_pdb_dict):
    """
    Produces a graph that relates clusters together

    Connections represent a PDB shared by 2 clusters. The more shared PDBs, the stronger the connection.
    """

    keys = list(cluster_pdb_dict.keys())
    keys_mapping = {l: k for l, k in enumerate(keys)}
    adjacency_matrix = np.zeros((len(keys), len(keys)))

    seen_dict = defaultdict(set)
    for i, key in enumerate(keys):
        for pdb in cluster_pdb_dict[key]:
            seen_dict[pdb].add(i)
    for cluster_set in seen_dict.values():
        for i, j in combinations(cluster_set, 2):
            adjacency_matrix[i, j] += 1
            adjacency_matrix[j, i] += 1

    graph = nx.from_numpy_matrix(adjacency_matrix)
    nx.relabel_nodes(graph, keys_mapping, copy=False)
    return graph


def _retrieve_seqs_names_list(merged_seqs_dict):
    """
    Retrieve all the sequences names that are contained in the merged_seqs_dict (same names as in the fasta file in input to MMSeqs2)
    """

    seqs_names = []
    for k in merged_seqs_dict.keys():
        for chain, _ in merged_seqs_dict[k]:
            seqs_names.append(k + "_" + chain)

    return seqs_names


def _create_pdb_seqs_dict(seqs_names_list):
    """
    Return a dictionary that has the PDB ids as keys and the list of all the chain names that correspond to this PDB
    """

    pdb_seqs_dict = defaultdict(lambda: [])
    for name in seqs_names_list:
        pdb_seqs_dict[name[:4]].append(name[5:])

    return pdb_seqs_dict


def _find_correspondances(files, dataset_dir):
    """
    Return a dictionary that contains all the biounits in the database (keys) and the list of all the chains that are in these biounits (values)
    """

    correspondances = defaultdict(lambda: [])
    for file in files:
        biounit = file
        with open(os.path.join(dataset_dir, file), "rb") as f:
            keys = pkl.load(f)
            for k in keys:
                correspondances[biounit].append(k)

    return correspondances


def _check_for_heteromers(grouped_seqs, biounit_chains):
    """
    True if the chain names contained in grouped_seqs correspond to at least 2 different sequences
    """

    grouped_seqs = [group.split("-") for group in grouped_seqs]
    chain_seqs = [
        np.argmax([chain in group for group in grouped_seqs])
        for chain in biounit_chains
    ]
    return np.max(chain_seqs) != np.min(chain_seqs)


def _divide_according_to_chains_interactions(pdb_seqs_dict, dataset_dir):
    """
    Divide all the biounit chains into 3 groups: single chains, homomers and heteromers, depending on the other chains present or not in the biounit
    """

    heteromers = []
    homomers = []
    single_chains = []
    all_files = np.array(list(os.listdir(dataset_dir)))
    all_pdb_files = np.array([f[:4] for f in all_files])

    for pdb in tqdm(pdb_seqs_dict.keys()):
        file_names = all_files[all_pdb_files == pdb]
        if type(file_names) == str:
            file_names = [file_names]
        seqs = pdb_seqs_dict[pdb]
        if len(seqs) == 1 and len(seqs[0].split("-")) == 1:
            for file_name in file_names:
                single_chains.append((file_name, seqs[0]))

        elif len(seqs) == 1 and len(file_names) == 1:
            for chain in seqs[0].split("-"):
                homomers.append((file_names[0], chain))

        elif len(seqs) == 1:
            correspondances = _find_correspondances(file_names, dataset_dir)
            for biounit in correspondances.keys():
                if len(correspondances[biounit]) == 1:
                    single_chains.append((biounit, correspondances[biounit][0]))
                else:
                    for chain in correspondances[biounit]:
                        homomers.append((biounit, chain))

        else:
            correspondances = _find_correspondances(file_names, dataset_dir)
            for biounit in correspondances.keys():
                if len(correspondances[biounit]) == 1:
                    single_chains.append((biounit, correspondances[biounit][0]))
                elif _check_for_heteromers(seqs, correspondances[biounit]):
                    for chain in correspondances[biounit]:
                        heteromers.append((biounit, chain))
                else:
                    for chain in correspondances[biounit]:
                        homomers.append((biounit, chain))

    return single_chains, homomers, heteromers


def _find_chains_in_graph(
    graph, clusters_dict, biounit_chains_array, pdbs_array, chains_array
):
    """
    Find all the biounit chains present in a given graph or subgraph

    Return a dictionary for which each key is a cluster name (merged chains name) and the values are all the biounit chains contained in this cluster.
    """

    res_dict = {}
    for k, node in enumerate(graph):
        grouped_chains = clusters_dict[node]
        split_chains = np.concatenate(
            [
                [(pdb[:4], chain) for chain in pdb[5:].split("-")]
                for pdb in grouped_chains
            ]
        )
        biounits_chains = np.concatenate(
            [
                biounit_chains_array[
                    np.logical_and(pdbs_array == pdb, chains_array == chain)
                ]
                for pdb, chain in split_chains
            ]
        )

        res_dict[node] = biounits_chains

    return res_dict


def _find_repartition(chains_dict, homomers, heteromers):
    """
    Return a dictionary similar to the one created by find_chains_in_graph, with an additional level of classification for single chains, homomers and heteromers

    Dictionary structure : `{'single_chains' : {cluster_name : [biounit chains]}, 'homomers' : {cluster_name : [biounit chains]}, 'heteromers' : {cluster_name : [biounit chains]}}`.
    Additionaly return the number of chains in each class (single chains, ...).
    """

    classes_dict = {
        "single_chains": defaultdict(lambda: []),
        "homomers": defaultdict(lambda: []),
        "heteromers": defaultdict(lambda: []),
    }
    n_single_chains, n_homomers, n_heteromers = 0, 0, 0

    for node in chains_dict.keys():
        for k, chain in enumerate(chains_dict[node]):
            if tuple(chain) in homomers:
                classes_dict["homomers"][node].append(chain)
                n_homomers += 1
            elif tuple(chain) in heteromers:
                classes_dict["heteromers"][node].append(chain)
                n_heteromers += 1
            else:
                classes_dict["single_chains"][node].append(chain)
                n_single_chains += 1

    for node in classes_dict["single_chains"]:
        classes_dict["single_chains"][node] = np.stack(
            classes_dict["single_chains"][node]
        )

    for node in classes_dict["homomers"]:
        classes_dict["homomers"][node] = np.stack(classes_dict["homomers"][node])

    for node in classes_dict["heteromers"]:
        classes_dict["heteromers"][node] = np.stack(classes_dict["heteromers"][node])

    return classes_dict, n_single_chains, n_homomers, n_heteromers


def _find_subgraphs_infos(
    subgraphs,
    clusters_dict,
    biounit_chains_array,
    pdbs_array,
    chains_array,
    homomers,
    heteromers,
):
    """
    Given a list of subgraphs, return a list of dictionaries and an array of sizes of the same length

    Dictionaries are the `chains_dict` and `classes_dict` corresponding to each subgraph, returned by the `find_chains_in_graph`
    and `find_repartition` functions respectively. The array of sizes is of shape (len(subgraph), 3). It gives the number of single chains,
    homomers and heteromers present in each subgraph.
    """

    size_array = np.zeros((len(subgraphs), 3))
    dict_list = []
    for k, subgraph in tqdm(enumerate(subgraphs)):
        chains_dict = _find_chains_in_graph(
            subgraph, clusters_dict, biounit_chains_array, pdbs_array, chains_array
        )
        classes_dict, n_single_chains, n_homomers, n_heteromers = _find_repartition(
            chains_dict, homomers, heteromers
        )
        size_array[k] = np.array([n_single_chains, n_homomers, n_heteromers])
        dict_list.append((chains_dict, classes_dict))

    return (
        dict_list,
        size_array,
        int(np.sum(size_array[:, 0])),
        int(np.sum(size_array[:, 1])),
        int(np.sum(size_array[:, 2])),
    )


def _construct_dataset(dict_list, size_array, indices):
    """
    Get a supergraph containing all subgraphs indicated by `indices`

    Given the `dict_list` and `size_array` returned by `find_subgraphs_info`, return the 2 dictionaries (`chains_dict` and `classes_dict`)
    corresponding to the graph encompassing all the subgraphs indicated by indices.
    Additionally return the number of single chains, homomers and heteromers in this supergraph.
    """

    dataset_clusters_dict = {}
    dataset_classes_dict = {"single_chains": {}, "homomers": {}, "heteromers": {}}
    single_chains_size, homomers_size, heteromers_size = 0, 0, 0
    for k in indices:
        chains_dict, classes_dict = dict_list[k]
        n_single_chains, n_homomers, n_heteromers = size_array[k]
        single_chains_size += n_single_chains
        homomers_size += n_homomers
        heteromers_size += n_heteromers

        dataset_clusters_dict.update(chains_dict)
        for chain_class in classes_dict.keys():
            dataset_classes_dict[chain_class].update(classes_dict[chain_class])

    return (
        dataset_clusters_dict,
        dataset_classes_dict,
        single_chains_size,
        homomers_size,
        heteromers_size,
    )


def _remove_elements_from_dataset(
    indices,
    remaining_indices,
    chain_class,
    size_obj,
    current_sizes,
    size_array,
    tolerance=0.2,
):
    """
    Remove values from indices untill we get the required (`size_obj`) number of chains in the class of interest (`chain_class`)

    Parameter `chain_class` corresponds to the single chain (0), homomer (1) or heteromer (2) class.
    """

    sizes = [s[chain_class] for s in size_array[indices]]
    sorted_sizes_indices = np.argsort(sizes)[::-1]

    while current_sizes[chain_class] > size_obj and len(sorted_sizes_indices) > 0:
        if (
            current_sizes[chain_class]
            - size_array[indices[sorted_sizes_indices[0]], chain_class]
            < (1 - tolerance) * size_obj
        ):
            sorted_sizes_indices = sorted_sizes_indices[1:]
            continue

        current_sizes -= size_array[indices[sorted_sizes_indices[0]]]
        remaining_indices.append(indices[sorted_sizes_indices[0]])
        indices.pop(sorted_sizes_indices[0])
        sizes = [s[chain_class] for s in size_array[indices]]
        sorted_sizes_indices = np.argsort(sizes)[::-1]

    return (
        indices,
        remaining_indices,
        current_sizes[0],
        current_sizes[1],
        current_sizes[2],
    )


def _check_mmseqs():
    """
    Raise an error if MMseqs2 is not installed
    """

    devnull = open(os.devnull, "w")
    retval = subprocess.call(
        ["mmseqs", "--help"], stdout=devnull, stderr=subprocess.STDOUT
    )
    devnull.close()
    if retval != 0:
        raise RuntimeError(
            "Please install the MMseqs2 library following the instructions at https://github.com/soedinglab/MMseqs2 (recommended: conda)"
        )


def _add_elements_to_dataset(
    indices,
    remaining_indices,
    chain_class,
    size_obj,
    current_sizes,
    size_array,
    tolerance=0.2,
):
    """
    Add values to indices untill we get the required (`size_obj`) number of chains in the class of interest (`chain_class`)

    Parameter `chain_class` corresponds to the single chain (0), homomer (1) or heteromer (2) class.
    """

    sizes = [s[chain_class] for s in size_array[remaining_indices]]
    sorted_sizes_indices = np.argsort(sizes)[::-1]

    while current_sizes[chain_class] < size_obj and len(sorted_sizes_indices) > 0:
        if (
            current_sizes[chain_class]
            + size_array[remaining_indices[sorted_sizes_indices[0]], chain_class]
            > (1 + tolerance) * size_obj
        ):
            sorted_sizes_indices = sorted_sizes_indices[1:]
            continue

        current_sizes += size_array[remaining_indices[sorted_sizes_indices[0]]].astype(
            int
        )
        indices.append(remaining_indices[sorted_sizes_indices[0]])
        remaining_indices.pop(sorted_sizes_indices[0])
        sizes = [s[chain_class] for s in size_array[remaining_indices]]
        sorted_sizes_indices = np.argsort(sizes)[::-1]

    return (
        indices,
        remaining_indices,
        current_sizes[0],
        current_sizes[1],
        current_sizes[2],
    )


def _adjust_dataset(
    indices,
    remaining_indices,
    dict_list,
    size_array,
    n_single_chains,
    n_homomers,
    n_heteromers,
    single_chains_size,
    homomers_size,
    heteromers_size,
    sc_available,
    hm_available,
    ht_available,
    tolerance=0.2,
):
    """
    If required, remove and add values in indices so that the number of chains in each class correspond to the required numbers within a tolerance

    First remove and then add (if necessary, for each class separately).
    In the end, we might end up with more chains than desired in the first 2 classes but for a reasonable tolerance (~10-20 %), this should not happen.
    """

    if single_chains_size > (1 + tolerance) * n_single_chains and sc_available:
        (
            indices,
            remaining_indices,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _remove_elements_from_dataset(
            indices,
            remaining_indices,
            0,
            n_single_chains,
            np.array([single_chains_size, homomers_size, heteromers_size]),
            size_array,
            tolerance=tolerance,
        )

    if homomers_size > (1 + tolerance) * n_homomers and hm_available:
        (
            indices,
            remaining_indices,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _remove_elements_from_dataset(
            indices,
            remaining_indices,
            1,
            n_homomers,
            np.array([single_chains_size, homomers_size, heteromers_size]),
            size_array,
            tolerance=tolerance,
        )

    if heteromers_size > (1 + tolerance) * n_heteromers and ht_available:
        (
            indices,
            remaining_indices,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _remove_elements_from_dataset(
            indices,
            remaining_indices,
            2,
            n_heteromers,
            np.array([single_chains_size, homomers_size, heteromers_size]),
            size_array,
            tolerance=tolerance,
        )

    if single_chains_size < (1 - tolerance) * n_single_chains and sc_available:
        (
            indices,
            remaining_indices,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _add_elements_to_dataset(
            indices,
            remaining_indices,
            0,
            n_single_chains,
            np.array([single_chains_size, homomers_size, heteromers_size]),
            size_array,
            tolerance=tolerance,
        )

    if homomers_size < (1 - tolerance) * n_homomers and hm_available:
        (
            indices,
            remaining_indices,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _add_elements_to_dataset(
            indices,
            remaining_indices,
            1,
            n_homomers,
            np.array([single_chains_size, homomers_size, heteromers_size]),
            size_array,
            tolerance=tolerance,
        )

    if heteromers_size < (1 - tolerance) * n_heteromers and ht_available:
        (
            indices,
            remaining_indices,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _add_elements_to_dataset(
            indices,
            remaining_indices,
            2,
            n_heteromers,
            np.array([single_chains_size, homomers_size, heteromers_size]),
            size_array,
            tolerance=tolerance,
        )

    (
        dataset_clusters_dict,
        dataset_classes_dict,
        single_chains_size,
        homomers_size,
        heteromers_size,
    ) = _construct_dataset(dict_list, size_array, indices)
    return (
        dataset_clusters_dict,
        dataset_classes_dict,
        single_chains_size,
        homomers_size,
        heteromers_size,
        remaining_indices,
    )


def test_availability(
    size_array,
    n_samples,
):
    """
    Test if there are enough groups in each class to construct a dataset with the required number of samples
    """

    present = np.sum(size_array != 0, axis=0)
    return present[0] > n_samples, present[1] > n_samples, present[2] > n_samples


def _fill_dataset(
    dict_list,
    size_array,
    n_samples,
    n_single_chains,
    n_homomers,
    n_heteromers,
    remaining_indices,
    n_max_iter=50,
    tolerance=0.2,
):
    """
    Construct a dataset from subgraphs indicated by `indices`

    Given a list of indices to choose from (`remaining_indices`), choose a list of subgraphs to construct a dataset containing the required number of
    biounits for each class (single chains, ...) within a tolerance.
    Return the same outputs as the construct_dataset function, as long as the list of remaining indices after selection.
    """

    single_chains_size, homomers_size, heteromers_size = 0, 0, 0
    sc_available, hm_available, ht_available = test_availability(
        size_array, n_samples
    )  # rule of thumb to estimate if it is logical to try to fill the dataset with a given class
    distribution_satisfied = False
    n_iter = 0

    while not distribution_satisfied and n_iter < n_max_iter:
        n_iter += 1
        indices = rd.sample(remaining_indices, n_samples)
        (
            dataset_clusters_dict,
            dataset_classes_dict,
            single_chains_size,
            homomers_size,
            heteromers_size,
        ) = _construct_dataset(dict_list, size_array, indices)
        distribution_satisfied = (
            (single_chains_size > (1 - tolerance) * n_single_chains or not sc_available)
            and (
                single_chains_size < (1 + tolerance) * n_single_chains
                or not sc_available
            )
            and (homomers_size > (1 - tolerance) * n_homomers or not hm_available)
            and (homomers_size < (1 + tolerance) * n_homomers or not hm_available)
            and (heteromers_size > (1 - tolerance) * n_heteromers or not ht_available)
            and (heteromers_size < (1 + tolerance) * n_heteromers or not ht_available)
        )

    if not distribution_satisfied:
        (
            dataset_clusters_dict,
            dataset_classes_dict,
            single_chains_size,
            homomers_size,
            heteromers_size,
            remaining_indices,
        ) = _adjust_dataset(
            indices,
            [i for i in remaining_indices if i not in indices],
            dict_list,
            size_array,
            n_single_chains,
            n_homomers,
            n_heteromers,
            single_chains_size,
            homomers_size,
            heteromers_size,
            sc_available,
            hm_available,
            ht_available,
            tolerance=tolerance,
        )
    else:
        remaining_indices = [i for i in remaining_indices if i not in indices]

    print("Number of samplings (fill_dataset):", n_iter)
    return (
        dataset_clusters_dict,
        dataset_classes_dict,
        remaining_indices,
        single_chains_size,
        homomers_size,
        heteromers_size,
    )


def _split_dataset(
    graph,
    clusters_dict,
    merged_seqs_dict,
    dataset_dir,
    valid_split=0.05,
    test_split=0.05,
    tolerance=0.2,
):
    """
    Given a graph representing connections between MMSeqs2 clusters, split the dataset between train, validation and test sets.
    Each onnected component of the graph is considered as a group.
    Then, groups are splitted into the 3 sets so that each set has the right amount of biounits.
    It has been observed that the biggest group represents about 15-20 % of all the biounits and thus it is automatically assigned to the train set.
    It is difficult to have the exact ratio of biounits in each set since biounits are manipulated by groups.
    However, within an acceptable tolerance (default 20 % of the split ratio - but in theory it can be smaller), the split ratios are respected.
    The process first try to randomly assign the groups to a set.
    If after 50 trials the partition fails to comply to the requirements, the last partition is kept and small adjustements are made by moving groups from sets one by one untill the ratios are approximately respected.
    Note that for very small datasets (around 250 biounits), this method will probably fail. But it also does not make much sense to use it for so few data.

    Parameters
    ----------
    graph : networkx graph
        the graph representing connections between MMSeqs2 clusters
    clusters_dict : dict
        the dictionary containing all the biounit files organized by clusters
    merged_seqs_dict : dict
        the dictionary containing all the merged (by similarity and PDB id) chains organized by cluster
    dataset_dir : str
        the path to the dataset
    valid_split : float, default 0.05
        the validation split ratio
    test_split : float, default 0.05
        the test split ratio

    Returns
    -------
    train_clusters_dict : dict
        the dictionary containing all the clusters (keys) and the biounit chains they contain for the training dataset
        structure : `{cluster_id : [(biounit_file_name, chain), (..., ...), ...]}`
    train_classes_dict : dict
        the same dictionary as train_cluster_dict with an additional level of classification (by single chains, homomers, heteromers)
        structure : `{'single_chains' : train_cluster_dict_like_dict, 'homomers' : train_cluster_dict_like_dict, 'heteromers' : train_cluster_dict_like_dict}`
    valid_clusters_dict : dict
        see train_clusters_dict but for validation set
    valid_classes_dict : dict
        see train_classes_dict but for validation set
    test_clusters_dict : dict
        see train_clusters_dict but for test set
    test_classes_dict : dict
        see train_classes_dict but for test set
    single_chains : list
        the list of all biounit chains (string names) that are in a single chain state (in their biounit)
    homomers : list
        the list of all biounit chains (string names) that are in a homomeric state (in their biounit)
    heteromers : list
        the list of all biounit chains (string names) that are in a heteromeric state (in their biounit)
    """

    subgraphs = np.array(
        [
            graph.subgraph(c)
            for c in sorted(nx.connected_components(graph), key=len, reverse=True)
        ],
        dtype=object,
    )
    remaining_indices = list(np.arange(1, len(subgraphs)))
    seqs_names_list = _retrieve_seqs_names_list(merged_seqs_dict)
    pdb_seqs_dict = _create_pdb_seqs_dict(seqs_names_list)
    single_chains, homomers, heteromers = _divide_according_to_chains_interactions(
        pdb_seqs_dict, dataset_dir
    )
    biounit_chains_array = np.array(single_chains + homomers + heteromers)
    pdbs_array = np.array([c[0][:4] for c in biounit_chains_array])
    chains_array = np.array([c[1] for c in biounit_chains_array])

    (
        dict_list,
        size_array,
        n_single_chains,
        n_homomers,
        n_heteromers,
    ) = _find_subgraphs_infos(
        subgraphs,
        clusters_dict,
        biounit_chains_array,
        pdbs_array,
        chains_array,
        homomers,
        heteromers,
    )

    (
        n_single_chains_valid,
        n_homomers_valid,
        n_heteromers_valid,
    ) = valid_split * np.array([n_single_chains, n_homomers, n_heteromers])
    n_single_chains_test, n_homomers_test, n_heteromers_test = test_split * np.array(
        [n_single_chains, n_homomers, n_heteromers]
    )
    n_samples_valid, n_samples_test = int(valid_split * len(subgraphs)), int(
        test_split * len(subgraphs)
    )

    (
        valid_clusters_dict,
        valid_classes_dict,
        remaining_indices,
        n_single_chains_valid,
        n_homomers_valid,
        n_heteromers_valid,
    ) = _fill_dataset(
        dict_list,
        size_array,
        n_samples_valid,
        n_single_chains_valid,
        n_homomers_valid,
        n_heteromers_valid,
        remaining_indices,
        tolerance=tolerance,
    )

    (
        test_clusters_dict,
        test_classes_dict,
        remaining_indices,
        n_single_chains_test,
        n_homomers_test,
        n_heteromers_test,
    ) = _fill_dataset(
        dict_list,
        size_array,
        n_samples_test,
        n_single_chains_test,
        n_homomers_test,
        n_heteromers_test,
        remaining_indices,
        tolerance=tolerance,
    )

    remaining_indices.append(
        0
    )  # add the big first cluster, that we always want in the training set
    (
        train_clusters_dict,
        train_classes_dict,
        n_single_chains_train,
        n_homomers_train,
        n_heteromers_train,
    ) = _construct_dataset(dict_list, size_array, remaining_indices)

    print("Classes distribution (single chain / homomer / heteromer):")
    print(
        "Train set:",
        int(n_single_chains_train),
        "/",
        int(n_homomers_train),
        "/",
        int(n_heteromers_train),
    )
    print(
        "Validation set:",
        int(n_single_chains_valid),
        "/",
        int(n_homomers_valid),
        "/",
        int(n_heteromers_valid),
    )
    print(
        "Test set:",
        int(n_single_chains_test),
        "/",
        int(n_homomers_test),
        "/",
        int(n_heteromers_test),
    )

    return (
        train_clusters_dict,
        train_classes_dict,
        valid_clusters_dict,
        valid_classes_dict,
        test_clusters_dict,
        test_classes_dict,
        single_chains,
        homomers,
        heteromers,
    )


def _build_dataset_partition(
    dataset_dir,
    tmp_folder,
    valid_split=0.05,
    test_split=0.05,
    tolerance=0.2,
    min_seq_id=0.3,
):
    """
    Build training, validation and test sets from a curated dataset of biounit, using MMSeqs2 for clustering

    Parameters
    ----------
    dataset_dir : str
        the path to the dataset
    valid_split : float in [0, 1], default 0.05
        the validation split ratio
    test_split : float in [0, 1], default 0.05
        the test split ratio
    min_seq_id : float in [0, 1], default 0.3
        minimum sequence identity for `mmseqs`

    Output
    ------
    train_clusters_dict : dict
        the dictionary containing all the clusters (keys) and the biounit chains they contain for the training dataset
        structure : {cluster_id : [(biounit_file_name, chain), (..., ...), ...]}
    train_classes_dict : dict
        the same dictionary as train_cluster_dict with an additional level of classification (by single chains, homomers, heteromers)
        structure : {'single_chains' : train_cluster_dict_like_dict, 'homomers' : train_cluster_dict_like_dict, 'heteromers' : train_cluster_dict_like_dict}
    valid_clusters_dict : dict
        see train_clusters_dict but for validation set
    valid_classes_dict : dict
        see train_classes_dict but for validation set
    test_clusters_dict : dict
        see train_clusters_dict but for test set
    test_classes_dict : dict
        see train_classes_dict but for test set
    """

    # retrieve all sequences and create a merged_seqs_dict
    merged_seqs_dict = _load_pdbs(dataset_dir)
    merged_seqs_dict = _merge_chains(merged_seqs_dict)

    # write sequences to a fasta file for clustering with MMSeqs2, run MMSeqs2 and delete the fasta file
    fasta_file = os.path.join(tmp_folder, "all_seqs.fasta")
    _write_fasta(fasta_file, merged_seqs_dict)
    _run_mmseqs2(fasta_file, tmp_folder, min_seq_id)
    subprocess.run(["rm", fasta_file])

    # retrieve MMSeqs2 clusters and build a graph with these clusters
    clusters_dict, clusters_pdb_dict = _read_clusters(
        os.path.join(tmp_folder, "MMSeqs2_results/clusterRes_all_seqs.fasta")
    )
    subprocess.run(["rm", "-r", os.path.join(tmp_folder, "MMSeqs2_results")])
    graph = _make_graph(clusters_pdb_dict)

    # perform the splitting into train, validation and tesst sets
    (
        train_clusters_dict,
        train_classes_dict,
        valid_clusters_dict,
        valid_classes_dict,
        test_clusters_dict,
        test_classes_dict,
        _,
        _,
        _,
    ) = _split_dataset(
        graph,
        clusters_dict,
        merged_seqs_dict,
        dataset_dir,
        valid_split=valid_split,
        test_split=test_split,
        tolerance=tolerance,
    )

    return (
        train_clusters_dict,
        train_classes_dict,
        valid_clusters_dict,
        valid_classes_dict,
        test_clusters_dict,
        test_classes_dict,
    )



# Adapted from: https://github.com/biotite-dev/biotite

from biotite.structure.geometry import distance, angle, dihedral
import numpy as np


def _annotate_sse(X):
    """
    Annotation of secondary structure elements (SSEs) in a chain
    """

    ca_coord = X[:, 2, :]
    length = ca_coord.shape[0]

    _r_helix = (np.deg2rad(89 - 12), np.deg2rad(89 + 12))
    _a_helix = (np.deg2rad(50 - 20), np.deg2rad(50 + 20))
    _d2_helix = ((5.5 - 0.5), (5.5 + 0.5))  # Not used in the algorithm description
    _d3_helix = ((5.3 - 0.5), (5.3 + 0.5))
    _d4_helix = ((6.4 - 0.6), (6.4 + 0.6))

    _r_strand = (np.deg2rad(124 - 14), np.deg2rad(124 + 14))
    _a_strand = (np.deg2rad(-180), np.deg2rad(-125), np.deg2rad(145), np.deg2rad(180))
    _d2_strand = ((6.7 - 0.6), (6.7 + 0.6))
    _d3_strand = ((9.9 - 0.9), (9.9 + 0.9))
    _d4_strand = ((12.4 - 1.1), (12.4 + 1.1))

    d2i = np.full(length, np.nan)
    d3i = np.full(length, np.nan)
    d4i = np.full(length, np.nan)
    ri = np.full(length, np.nan)
    ai = np.full(length, np.nan)

    d2i[1 : length - 1] = distance(ca_coord[0 : length - 2], ca_coord[2:length])
    d3i[1 : length - 2] = distance(ca_coord[0 : length - 3], ca_coord[3:length])
    d4i[1 : length - 3] = distance(ca_coord[0 : length - 4], ca_coord[4:length])
    ri[1 : length - 1] = angle(
        ca_coord[0 : length - 2], ca_coord[1 : length - 1], ca_coord[2:length]
    )
    ai[1 : length - 2] = dihedral(
        ca_coord[0 : length - 3],
        ca_coord[1 : length - 2],
        ca_coord[2 : length - 1],
        ca_coord[3 : length - 0],
    )

    sse = np.full(len(ca_coord), "c", dtype="U1")
    # Annotate helices
    # Find CA that meet criteria for potential helices
    is_pot_helix = np.zeros(len(sse), dtype=bool)
    for i in range(len(sse)):
        if (
            d3i[i] >= _d3_helix[0]
            and d3i[i] <= _d3_helix[1]
            and d4i[i] >= _d4_helix[0]
            and d4i[i] <= _d4_helix[1]
        ) or (
            ri[i] >= _r_helix[0]
            and ri[i] <= _r_helix[1]
            and ai[i] >= _a_helix[0]
            and ai[i] <= _a_helix[1]
        ):
            is_pot_helix[i] = True
    # Real helices are 5 consecutive helix elements
    is_helix = np.zeros(len(sse), dtype=bool)
    counter = 0
    for i in range(len(sse)):
        if is_pot_helix[i]:
            counter += 1
        else:
            if counter >= 5:
                is_helix[i - counter : i] = True
            counter = 0
    # Extend the helices by one at each end if CA meets extension criteria
    i = 0
    while i < len(sse):
        if is_helix[i]:
            sse[i] = "a"
            if (d3i[i - 1] >= _d3_helix[0] and d3i[i - 1] <= _d3_helix[1]) or (
                ri[i - 1] >= _r_helix[0] and ri[i - 1] <= _r_helix[1]
            ):
                sse[i - 1] = "a"
            sse[i] = "a"
            if (d3i[i + 1] >= _d3_helix[0] and d3i[i + 1] <= _d3_helix[1]) or (
                ri[i + 1] >= _r_helix[0] and ri[i + 1] <= _r_helix[1]
            ):
                sse[i + 1] = "a"
        i += 1

    # Annotate sheets
    # Find CA that meet criteria for potential strands
    is_pot_strand = np.zeros(len(sse), dtype=bool)
    for i in range(len(sse)):
        if (
            d2i[i] >= _d2_strand[0]
            and d2i[i] <= _d2_strand[1]
            and d3i[i] >= _d3_strand[0]
            and d3i[i] <= _d3_strand[1]
            and d4i[i] >= _d4_strand[0]
            and d4i[i] <= _d4_strand[1]
        ) or (
            ri[i] >= _r_strand[0]
            and ri[i] <= _r_strand[1]
            and (
                (ai[i] >= _a_strand[0] and ai[i] <= _a_strand[1])
                or (ai[i] >= _a_strand[2] and ai[i] <= _a_strand[3])
            )
        ):
            is_pot_strand[i] = True
    # Real strands are 5 consecutive strand elements,
    # or shorter fragments of at least 3 consecutive strand residues,
    # if they are in hydrogen bond proximity to 5 other residues
    is_strand = np.zeros(len(sse), dtype=bool)
    counter = 0
    contacts = 0
    for i in range(len(sse)):
        if is_pot_strand[i]:
            counter += 1
            coord = ca_coord[i]
            for strand_coord in ca_coord:
                dist = distance(coord, strand_coord)
                if dist >= 4.2 and dist <= 5.2:
                    contacts += 1
        else:
            if counter >= 4:
                is_strand[i - counter : i] = True
            elif counter == 3 and contacts >= 5:
                is_strand[i - counter : i] = True
            counter = 0
            contacts = 0
    # Extend the strands by one at each end if CA meets extension criteria
    i = 0
    while i < len(sse):
        if is_strand[i]:
            sse[i] = "b"
            if d3i[i - 1] >= _d3_strand[0] and d3i[i - 1] <= _d3_strand[1]:
                sse[i - 1] = "b"
            sse[i] = "b"
            if d3i[i + 1] >= _d3_strand[0] and d3i[i + 1] <= _d3_strand[1]:
                sse[i + 1] = "b"
        i += 1

    return sse



from proteinflow import generate_data, split_data
import os
import pickle
from collections import defaultdict
import editdistance
import shutil
from time import time
import pytest


def get_class(seqs_dict):
    """Check if the protein is a homomer, a heteromer or a single chain"""

    keys = list(seqs_dict.keys())
    if len(keys) == 1:
        return "single_chains"
    ref_seq = list(seqs_dict.values())[0]["seq"]
    for key in keys[1:]:
        value = seqs_dict[key]
        seq = value["seq"]
        if (
            len(seq) > 1.1 * len(ref_seq)
            or len(seq) < 0.9 * len(ref_seq)
            or editdistance.eval(seq, ref_seq) / max(len(seq), len(ref_seq)) > 0.1
        ):
            return "heteromers"
    return "homomers"


# @pytest.mark.skip()
def test_generate():
    """Test generate_data + split_data + chain class distribution"""

    folder = "./data/proteinflow_test"
    if os.path.exists(folder):
        shutil.rmtree(folder)
    start = time()
    generate_data(tag="test", skip_splitting=True, n=50)
    end = time()
    num_files = len(os.listdir(folder))
    split_data(tag="test", valid_split=0.2, test_split=0.1)
    assert os.path.exists(folder)
    assert len(os.listdir(folder)) == 4
    num_files_split = 0

    for subset in ["train", "valid", "test"]:
        num_files_split += len(os.listdir(os.path.join(folder, subset)))
        subset_folder = os.path.join(folder, subset)
        with open(os.path.join(folder, "splits_dict", f"{subset}.pickle"), "rb") as f:
            for _ in range(2):
                class_data = pickle.load(f)
        classes = defaultdict(int)
        for file in os.listdir(subset_folder):
            with open(os.path.join(subset_folder, file), "rb") as f:
                data = pickle.load(f)
            c = get_class(data)
            classes[c] += 1
        for c in class_data:
            class_files = set()
            for chain_arr in class_data[c].values():
                for file, _ in chain_arr:
                    class_files.add(file)
            assert classes[c] == len(class_files)
    assert num_files == num_files_split
    shutil.rmtree(folder)
    print(f"generation time: {end - start} sec")


if __name__ == "__main__":
    test_generate()
